// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file DataCollector.h */

#ifndef AMDIS_DATACOLLECTOR_H
#define AMDIS_DATACOLLECTOR_H

#include <list>
#include <vector>

#include "ElementInfo.h"
#include "VertexInfo.h"
#include "PeriodicInfo.h"
#include "MemoryManager.h"
#include "Mesh.h"

namespace AMDiS {
  class VertexInfo;
  class ElementInfo;
  class Mesh;
  class FiniteElemSpace;
  template<typename T> class DOFVector;

  /** 
   * \ingroup Output
   *
   * \brief
   * Provides data collecting of element, vertex and value data 
   * for file writers.
   */
  class DataCollector 
    {
    public:
      MEMORY_MANAGED(DataCollector);

      /** \brief
       * Constructor.
       */
      DataCollector(const FiniteElemSpace *feSpace,
		    DOFVector<double> *values,
		    int level = -1,
		    Flag traverseFlag = Mesh::CALL_LEAF_EL,
		    bool (*writeElem)(ElInfo*) = NULL);

      ~DataCollector();

      /** \brief
       * Returns list with element information.
       */
      ::std::list<ElementInfo>* getElementInfos();

      /** \brief
       * Returns vector with vertex information.
       */
      DOFVector< ::std::list<VertexInfo> >* getVertexInfos();

      /** \brief
       * Returns the finite element space of the problem.
       */
      const FiniteElemSpace* getFeSpace();

      /** \brief
       * Returns vector with value information.
       */     
      DOFVector<double>* getValues();

      /** \brief
       * Returns vector with information about dof coordinates.
       */     
      DOFVector< ::std::list<WorldVector<double> > >* getDofCoords();

      /** \brief
       * Returns vector with information about interpolation 
       * point indexing.
       */          
      DOFVector<int>* getInterpPointInd();

      /** \brief
       *
       */
      DOFVector< ::std::list<WorldVector<double> > >* getInterpPointCoords();

      /** \brief
       * Returns list of interpolation point information.
       */
      ::std::vector< ::std::vector<int> >* getInterpPoints();

      /** \brief
       * Returns list of information about periodics.
       */
      ::std::list<PeriodicInfo>* getPeriodicInfos();

      /** \brief
       * Returns the number of vertices.
       */
      int getNumberVertices();

      /** \brief
       * Returns the number of elements.
       */
      int getNumberElements();

      /** \brief
       * Returns the number of interpolation points.
       */
      int getNumberInterpPoints();

      /** \brief
       * Returns the number of connections.
       */
      int getNumberConnections();
      
      /** \brief
       * Returns the mesh of the problem.
       */
      Mesh* getMesh();

    protected:  
      /** \brief
       * Start collecting element and vertex data of the problem.
       */
      int startCollectingElementData();

      /** \brief
       * Start collecting value data of the problem.
       */
      int startCollectingValueData();

      /** \brief
       * Start collecting periodic data of the problem.
       */
      int startCollectingPeriodicData();

      /** \brief
       * Adds information about one element and its vertices.
       */
      int addElementData(ElInfo* elInfo);

      /** \brief
       * Adds value information of one element.
       */
      int addValueData(ElInfo *elInfo);

      /** \brief
       * Adds information about interpolation points of vertices.
       */
      int addInterpData(ElInfo *elInfo);

      /** \brief
       * Adds value information of one element.
       */
      int addPeriodicData(ElInfo *elInfo);


      /** \brief
       * Vector with vertex values
       */
      DOFVector<double> *values_;

      /** \brief
       * Level information for traversing the mesh.
       */
      int level_;

      /** \brief
       * Flags for traversing the mesh.
       */
      Flag traverseFlag_;

      /** \brief
       * 
       */
      const FiniteElemSpace *feSpace_;

      /** \brief
       * Mesh that should be written
       */
      Mesh *mesh_;
      
      /** \brief
       * DOFAdmin of values
       */
      DOFAdmin *localAdmin_;

      /** \brief
       * vertex pre-dofs
       */
      int nPreDofs_;

      /** \brief
       * Number of vertices.
       */
      int nVertices_;

      /** \brief
       * Number of elements.
       */
      int nElements_;

      /** \brief
       * Total number of interpolation points.
       */
      int nInterpPoints_;

      /** \brief
       * Number of connections in periodic problems.
       */
      int nConnections_;

      /** \brief
       * Dimension of \ref mesh
       */
      int dim_;
           
      /** \brief
       * Maps internal element indices to global output indices.
       */
      ::std::map<int, int> outputIndices_;

      /** \brief
       * Global interpolation point indexing
       */
      DOFVector<int> *interpPointInd_;

      /** \brief
       * Stores for each simplex the interpolation points.
       */
      ::std::vector< ::std::vector<int> > interpPoints_;

      /** \brief
       *
       */
      DOFVector< ::std::list<WorldVector<double> > > *interpPointCoords_;

      /** \brief
       * list of coords for each dof
       */
      DOFVector< ::std::list<WorldVector<double> > > *dofCoords_;

      /** \brief
       * List that stores an ElementInfo for each element.
       */
      ::std::list<ElementInfo> elements_;

      /** \brief
       * List stat stores information about all periodics.
       */
      ::std::list<PeriodicInfo> periodicInfos_;

      /** \brief
       * Stores a list of vertex infos for each dof.
       */
      DOFVector< ::std::list<VertexInfo> > *vertexInfos_;

      /** \brief
       * periodicConnections[i][j] stores whether the connection at side j of 
       * the element with output index i has already been written.
       */
      ::std::vector<DimVec<bool> > periodicConnections_;

      /** \brief
       * Stores if element data was collected before.
       */
      bool elementDataCollected_;

      /** \brief
       * Stores if value data was collected before.
       */     
      bool valueDataCollected_;

      /** \brief
       * Stores if periodic data was collected before.
       */     
      bool periodicDataCollected_;

      /** \brief
       * Pointer to a function which decides whether an element is considered.
       */
      bool (*writeElem_)(ElInfo*);
    };
}

#endif
